/*
 * This file is part of the ToolBox.
 * Copyright Thomas Jacob.
 *
 * READ README.TXT BEFORE USE!!
 */


#ifndef __TOOLBOX_CONFIG_H
#define __TOOLBOX_CONFIG_H


#define _CF_TOKENTYPE_OPENING 1
#define _CF_TOKENTYPE_CLOSING 2
#define _CF_TOKENTYPE_PLAIN 3
#define _CF_TOKENTYPE_IGNORE 4


/**
 * The ConfigParameter is not yet assigned.
 * This constant is internally used.
 */
#define CF_TNONE   0

/**
 * The ConfigParameter is a boolean.
 * This constant is internally used.
 */
#define CF_TBOOL   1

/**
 * The ConfigParameter is a double precision float.
 * This constant is internally used.
 */
#define CF_TDOUBLE 2

/**
 * The ConfigParameter is a long integer.
 * This constant is internally used.
 */
#define CF_TLONG   3

/**
 * The ConfigParameter is a string.
 * This constant is internally used.
 */
#define CF_TSTRING 4

/**
 * The string writer buffer increase amount if it is full.
 * This constant is internally used.
 */
#define CF_SWBUFFERINCREASE 512


namespace toolbox
{
	/**
	 * The Config class stores information about the
	 * application that is user editable or about to change.
	 * This configuration is structured into sections,
	 * containing sections and/or parameters. The sections
	 * may be nested as deep as required.
	 * @author Thomas Jacob
	 */
	#if defined(_DEBUG) && (defined(_AFX) || defined(_AFXDLL))
		class Config : public CObject
	#else
		class Config
	#endif
	{
		friend ConfigSection;
		friend ConfigParameter;

		public:

			/**
			 * The super class for the writing targets. This class is used
			 * to provide different targets like files and strings.
			 */
			#if defined(_DEBUG) && (defined(_AFX) || defined(_AFXDLL))
				class Writer : public CObject
			#else
				class Writer
			#endif
			{
				public:

					/**
					 * Destroys the writer.
					 */
					virtual ~Writer();

					/**
					 * Writes a character to the writer's target.
					 * @param character The character to be written.
					 * @return Whether the writing was successful.
					 * @note Implement this method in derivatives.
					 */
					virtual bool Write(char character) = NULL;

					/**
					 * Writes a string to the writer's target.
					 * @param string The string to be written.
					 * @return Whether the writing was successful.
					 * @note Implement this method in derivatives.
					 */
					virtual bool Write(const char * string) = NULL;
			};

			/**
			 * The file writer is a writer that writes to a file.
			 */
			class FileWriter : public Writer
			{
				private:

					/**
					 * The file to write to.
					 */
					FILE * File;

				public:

					/**
					 * Creates a new file writer.
					 */
					FileWriter();

					/**
					 * Destroys the file writer.
					 * This also closes the open file, if any.
					 */
					virtual ~FileWriter();

					/**
					 * Opens a file for the writer.
					 * The method fails if you try this twice.
					 * @param fileName The name of the file to be opened.
					 * @return Whether the file could be opened.
					 */
					bool Open(const char * fileName);

					/**
					 * Writes a character to the writer's target.
					 * @param character The character to be written.
					 * @return Whether the writing was successful.
					 * @note Implement this method in derivatives.
					 */
					virtual bool Write(char character);

					/**
					 * Writes a string to the writer's target.
					 * @param string The string to be written.
					 * @return Whether the writing was successful.
					 * @note Implement this method in derivatives.
					 */
					virtual bool Write(const char * string);
			};

			/**
			 * The string writer is a writer that creates a string.
			 */
			class StringWriter : public Writer
			{
				private:

					/**
					 * The so far accumulated string.
					 */
					char * Buffer;

					/**
					 * The current length of the buffer.
					 * This is the number of used (filled) bytes.
					 */
					unsigned int BufferLen;

					/**
					 * The current size of the buffer.
					 * This is the number of reserved (allocated) bytes.
					 */
					unsigned int BufferSize;

					/**
					 * Increases the buffer's size by CF_SWBUFFERINCREASE bytes.
					 */
					void IncreaseBuffer();

				public:

					/**
					 * Creates a new string writer.
					 */
					StringWriter();

					/**
					 * Destroys the string writer.
					 */
					virtual ~StringWriter();

					/**
					 * Returns the so far accumulated string.
					 * @return The string.
					 * @warning You must not delete nor modify this string.
					 * It is automatically deleted when destroying the writer.
					 */
					inline const char * GetString();

					/**
					 * Writes a character to the writer's target.
					 * @param character The character to be written.
					 * @return Whether the writing was successful.
					 * @note Implement this method in derivatives.
					 */
					virtual bool Write(char character);

					/**
					 * Writes a string to the writer's target.
					 * @param string The string to be written.
					 * @return Whether the writing was successful.
					 * @note Implement this method in derivatives.
					 */
					virtual bool Write(const char * string);
			};

		private:

			/**
			 * The main section, the root entry for all
			 * further sections and parameters.
			 */
			ConfigSection * Content;

			#ifdef _TOOLBOX_TEST
				/**
				 * The number of existing instances for testing purposes.
				 */
				static int InstanceCount;
			#endif

			/**
			 * Writes a given string to a file, after having it
			 * modified so that all illegal characters are escaped
			 * with a back slash.
			 * @param Writer The writer to write to.
			 * @param str The string to be written.
			 * @return Whether the writing was successful.
			 */
			static bool WriteCorrectedForCfgFile(class Config::Writer * Writer, const char * str);

		public:

			/**
			 * Creates a new configuration.
			 */
			Config();

			/**
			 * Destroys the configuration.
			 */
			~Config();

			/**
			 * Clears the whole configuration object,
			 * removing all sections and parameters.
			 */
			void Clear();

			/**
			 * Exports the configuration into an INI file.
			 * The method uses the application's name with the suffix INI.
			 * Use the parametered overload to choose the file name.
			 * @warning Since INI files do not support nested sections,
			 * any sections deeper than 1 will be lost.
			 * @return If the configuration was successfully written.
			 */
			bool ExportIni();

			/**
			 * Exports the configuration into an INI file.
			 * @param fileName The name of the file to write to.
			 * @warning Since INI files do not support nested sections,
			 * any sections deeper than 1 will be lost.
			 * @return If the configuration was successfully written.
			 */
			bool ExportIni(const char * fileName);

			/**
			 * Returns the main configuration section, the root entry.
			 * @return The main section.
			 */
			inline ConfigSection * GetContent();

			/**
			 * Imports an INI file as configuration.
			 * Any configuration in the object is deleted in prior.
			 * The method uses the application's name with the suffix INI.
			 * Use the parametered overload to choose the file name.
			 * @return If the configuration was successfully read.
			 * This value is false, especially when the file does not
			 * exist or when it has syntax errors. In this case, the
			 * configuration remains empty.
			 */
			bool ImportIni();

			/**
			 * Imports an INI file as configuration.
			 * Any configuration in the object is deleted in prior.
			 * @param fileName The name of the file to read from.
			 * @return If the configuration was successfully read.
			 * This value is false, especially when the file does not
			 * exist or when it has syntax errors. In this case, the
			 * configuration remains empty.
			 */
			bool ImportIni(const char * fileName);

			/**
			 * Deserializes (reads) the configuration from a string.
			 * @param text The string to read the config from.
			 * @return If the configuration was successfully read.
			 * This value is false, if the string is syntactically
			 * invalid. In this case, the configuration remains empty.
			 */
			bool Parse(char * text);

			/**
			 * Reads the configuration from a config file.
			 * Any configuration in the object is deleted in prior.
			 * The method uses the application's name with the suffix CFG.
			 * Use the parametered overload to choose the file name.
			 * @return If the configuration was successfully read.
			 * This value is false, especially when the file does not
			 * exist or when it has syntax errors. In this case, the
			 * configuration remains empty.
			 */
			bool Read();

			/**
			 * Reads the configuration from a config file.
			 * Any configuration in the object is deleted in prior.
			 * @param fileName The name of the file to read from.
			 * @return If the configuration was successfully read.
			 * This value is false, especially when the file does not
			 * exist or when it has syntax errors. In this case, the
			 * configuration remains empty.
			 */
			bool Read(const char * fileName);

			#ifdef _TOOLBOX_TEST
				/**
				 * Performs a series of tests to check if the class
				 * runs properly. Produces standard output on errors.
				 * @param performedTests Is increased by one for each performed test.
				 * @param failedTests Is increased by one for each failed test.
				 */
				static void RunTestSuite(int * performedTests, int * failedTests);
			#endif

			/**
			 * Writes the configuration into a config file.
			 * The method uses the application's name with the suffix CFG.
			 * Use the parametered overload to choose the file name.
			 * @param withBacking Whether the configuration shall be written
			 * to a backing file first, and after writing has been completed,
			 * it shall be moved to its final destination. Default is false.
			 * @param indent Whether to indent values according to their
			 * structural depth. Default is false.
			 * @return If the configuration was successfully written.
			 */
			bool Write(bool withBacking = false, bool indent = true);

			/**
			 * Writes the configuration into a config file.
			 * @param fileName The name of the file to write to.
			 * @param withBacking Whether the configuration shall be written
			 * to a backing file first, and after writing has been completed,
			 * it shall be moved to its final destination. Default is false.
			 * @param indent Whether to indent values according to their
			 * structural depth. Default is false.
			 * @return If the configuration was successfully written.
			 */
			bool Write(const char * fileName, bool withBacking = false, bool indent = true);

			/**
			 * Writes the configuration into a given writer, for example a string writer.
			 * @param Writer The writer to write to.
			 * @param indent Whether to indent values according to their
			 * structural depth. Default is false.
			 * @return Whether the configuration was successfully written.
			 */
			bool Write(class Config::Writer * Writer, bool indent = true);
	};


	/**
	 * A ConfigSection stores all sub-sections and parameters
	 * of a section in the configuation.
	 * @author Thomas Jacob
	 */
	#if defined(_DEBUG) && (defined(_AFX) || defined(_AFXDLL))
		class ConfigSection : public CObject
	#else
		class ConfigSection
	#endif
	{
		friend Config;
		friend HashMap<ConfigSection>;

		private:

			#ifdef _TOOLBOX_TEST
				/**
				 * The number of existing instances for testing purposes.
				 */
				static int InstanceCount;
			#endif

			/**
			 * The name of the section.
			 */
			char * Name;

			/**
			 * The parameters in this section.
			 */
			StringKeyHashMap<ConfigParameter> * Parameters;

			/**
			 * The sub-sections in this section.
			 */
			StringKeyHashMap<ConfigSection> * Sections;

			/**
			 * Creates a new config section.
			 * @param name The name of the section.
			 */
			ConfigSection(const char * name);

			/**
			 * Destroys the config section.
			 */
			~ConfigSection();

			/**
			 * Writes the indentation (tabs) according to the given
			 * depth into a file.
			 * @param Writer The writer to write to.
			 * @param depth The number of tabs to be written.
			 * @return Whether the writing was successful.
			 */
			static bool WriteIndentation(class Config::Writer * Writer, int depth);

			/**
			 * Writes this section to a file.
			 * @param Writer The writer to write to.
			 * @param depth The section nest depth, to indent correctly.
			 * @param indent Whether to indent values according to their
			 * structural depth.
			 * @return Whether the writing was successful.
			 */
			bool WriteTo(class Config::Writer * Writer, int depth, bool indent);

		public:

			/**
			 * Returns whether a given parameter is existing.
			 * @param name The name of the parameter.
			 * @return Whether the parameter is existing.
			 */
			bool ContainsParameter(const char * name);

			/**
			 * Returns whether a given sub-section is existing.
			 * @param name The name of the section.
			 * @return Whether the section is existing.
			 */
			bool ContainsSection(const char * name);

			/**
			 * Removes all parameters from the section.
			 */
			void DeleteAllParameters();

			/**
			 * Removes all sub-sections from the section.
			 */
			void DeleteAllSections();

			/**
			 * Removes a parameter from the section.
			 * @param name The name of the parameter.
			 */
			void DeleteParameter(const char * name);

			/**
			 * Removes a sub-section from the section.
			 * @param name The name of the sub-section.
			 */
			void DeleteSection(const char * name);

			/**
			 * Returns a boolean value for a given name.
			 * If the parameter is not of boolean type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @return The boolean value.
			 */
			inline bool GetBoolValue(const char * name);

			/**
			 * Returns a boolean value of a parameter.
			 * If the parameter is not of boolean type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @param defaultBoolValue The value to be used, if
			 * the parameter did not exist.
			 * @return The boolean value.
			 */
			bool GetBoolValue(const char * name, bool defaultBoolValue);

			/**
			 * Returns a double float value of a parameter.
			 * If the parameter is not of double float type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @return The double float value.
			 */
			inline double GetDoubleValue(const char * name);

			/**
			 * Returns a double float value of a parameter.
			 * If the parameter is not of double float type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @param defaultDoubleValue The value to be used, if
			 * the parameter did not exist.
			 * @return The double float value.
			 */
			double GetDoubleValue(const char * name, double defaultDoubleValue);

			/**
			 * Returns a integer value of a parameter.
			 * If the parameter is not of integer type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @return The integer value.
			 * @warning Since the long type is also used
			 * for ints, this might mean a loss of leading digits.
			 */
			inline int GetIntValue(const char * name);

			/**
			 * Returns a integer value of a parameter.
			 * If the parameter is not of integer type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @param defaultIntValue The value to be used, if
			 * the parameter did not exist.
			 * @return The integer value.
			 * @warning Since the long type is also used
			 * for ints, this might mean a loss of leading digits.
			 */
			int GetIntValue(const char * name, int defaultIntValue);

			/**
			 * Returns a long integer value of a parameter.
			 * If the parameter is not of long integer type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @return The long integer value.
			 */
			inline long GetLongValue(const char * name);

			/**
			 * Returns a long integer value of a parameter.
			 * If the parameter is not of long integer type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @param name The name of the parameter.
			 * @param defaultLongValue The value to be used, if
			 * the parameter did not exist.
			 * @return The long integer value.
			 */
			long GetLongValue(const char * name, long defaultLongValue);

			/**
			 * Returns the name of the section.
			 * @return The name of the section.
			 */
			inline const char * GetName();

			/**
			 * Returns the ConfigParameter object for a given name
			 * or creates it, if it doesn't exist yet.
			 * @param name The name of the parameter.
			 * @return The ConfigParameter.
			 */
			ConfigParameter * GetParameter(const char * name);

			/**
			 * Returns a PointeredList of all parameter names
			 * in this section.
			 * @return The parameter names as PointeredList.
			 * @warning This method creates a PointeredList and
			 * expects that it is deleted by the caller.
			 */
			inline PointeredList * GetParameterNames();

			/**
			 * Returns a sub-section of the section. If the
			 * section does not exist, a new ConfigSection is returned.
			 * So use this method to create sections as well.
			 * @param name The name of the section.
			 * @return The section according to the name.
			 */
			ConfigSection * GetSection(const char * name);

			/**
			 * Returns a PointeredList of all section names
			 * in this section.
			 * @return The section names as PointeredList.
			 * @warning This method creates a PointeredList and
			 * expects that it is deleted by the caller.
			 */
			inline PointeredList * GetSectionNames();

			/**
			 * Returns a PointeredList of all sections
			 * in this section.
			 * @return The sections as PointeredList.
			 * @warning This method creates a PointeredList and
			 * expects that it is deleted by the caller.
			 */
			inline PointeredList * GetSections();

			/**
			 * Returns a string value of a parameter.
			 * If the parameter is not of string type,
			 * it is converted in prior, as far as this does
			 * not mean any memory allocation. If the conversion
			 * isn't possible without allocating memory (for
			 * example for the integer type), the method
			 * returns an empty string.
			 * @param name The name of the parameter.
			 * @return The string value.
			 * @note Do not free or change the return value.
			 * Use SetStringValue instead.
			 */
			inline const char * GetStringValue(const char * name);

			/**
			 * Returns a string value of a parameter.
			 * If the parameter is not of string type,
			 * it is converted in prior, as far as this does
			 * not mean any memory allocation. If the conversion
			 * isn't possible without allocating memory (for
			 * example for the integer type), the method
			 * returns an empty string.
			 * @param name The name of the parameter.
			 * @param defaultStringValue The value to be used, if
			 * the parameter did not exist. Its memory is copied,
			 * it is no longer used after the call.
			 * @return The string value.
			 * @note Do not free or change the return value.
			 * Use SetStringValue instead.
			 */
			const char * GetStringValue(const char * name, const char * defaultStringValue);

			/**
			 * Returns the type of a given parameter.
			 * @param name The name of the parameter.
			 * @return The type of the parameter.
			 */
			inline int GetType(const char * name);

			/**
			 * Returns whether the config section has a parameter with a given name.
			 * @param name The name of the parameter to be checked.
			 * @return Whether the config section has the parameter.
			 */
			bool HasParameter(const char * name);

			/**
			 * Returns whether the config section has a sub section with a given name.
			 * @param name The name of the sub section to be checked.
			 * @return Whether the config section has the sub section.
			 */
			bool HasSection(const char * name);

			#ifdef _TOOLBOX_TEST
				/**
				 * Performs a series of tests to check if the class
				 * runs properly. Produces standard output on errors.
				 * @param performedTests Is increased by one for each performed test.
				 * @param failedTests Is increased by one for each failed test.
				 */
				static void RunTestSuite(int * performedTests, int * failedTests);
			#endif

			/**
			 * Sets a boolean parameter for a given name.
			 * @param name The name of the parameter.
			 * @param boolValue The new boolean value.
			 */
			inline void SetBoolValue(const char * name, bool boolValue);

			/**
			 * Sets a double float parameter for a given name.
			 * @param name The name of the parameter.
			 * @param doubleValue The new double float value.
			 */
			inline void SetDoubleValue(const char * name, double doubleValue);

			/**
			 * Sets a integer parameter for a given name.
			 * @param name The name of the parameter.
			 * @param intValue The new integer value.
			 * @note This method is provided for better
			 * compatibility of the int and long types.
			 */
			inline void SetIntValue(const char * name, int intValue);

			/**
			 * Sets a long integer parameter for a given name.
			 * @param name The name of the parameter.
			 * @param longValue The new long integer value.
			 */
			inline void SetLongValue(const char * name, long longValue);

			/**
			 * Sets a string parameter for a given name.
			 * @param name The name of the parameter.
			 * @param stringValue The new string value.
			 * The method copies the string so its memory
			 * is not used after the creation.
			 */
			inline void SetStringValue(const char * name, const char * stringValue);
	};


	/**
	 * A ConfigParameter is the single value in the
	 * configuration. It can be stored in the main section
	 * or a sub-section.
	 * @author Thomas Jacob
	 */
	#if defined(_DEBUG) && (defined(_AFX) || defined(_AFXDLL))
		class ConfigParameter : public CObject
	#else
		class ConfigParameter
	#endif
	{
		friend ConfigSection;
		friend HashMap<ConfigParameter>;

		private:

			union
			{
				/**
				 * The boolean value of the parameter,
				 * if the parameter's type is boolean.
				 */
				bool BoolValue;

				/**
				 * The double float value of the parameter,
				 * if the parameter's type is double float.
				 */
				double DoubleValue;

				/**
				 * The long integer value of the parameter,
				 * if the parameter's type is long integer.
				 */
				long LongValue;

				/**
				 * The string (char *) value of the parameter,
				 * if the parameter's type is string.
				 */
				char * StringValue;
			};

			#ifdef _TOOLBOX_TEST
				/**
				 * The number of existing instances for testing purposes.
				 */
				static int InstanceCount;
			#endif

			/**
			 * The current type of the parameter.
			 */
			int Type;

			/**
			 * Creates a config parameter.
			 */
			ConfigParameter();

			/**
			 * Creates a boolean config parameter.
			 * @param boolValue The boolean value.
			 */
			ConfigParameter(bool boolValue);

			/**
			 * Creates a double float config parameter.
			 * @param doubleValue The double float value.
			 */
			ConfigParameter(double doubleValue);

			/**
			 * Creates a long integer config parameter.
			 * @param longValue The long integer value.
			 */
			ConfigParameter(long longValue);

			/**
			 * Creates a string (char *) config parameter.
			 * @param stringValue The string value.
			 * @note The method copies the string so its memory
			 * is not used after the creation.
			 */
			ConfigParameter(const char * stringValue);

			/**
			 * Destroys the config parameter.
			 */
			~ConfigParameter();

			/**
			 * Writes the parameter to a file, after having it
			 * modified so that all illegal characters are escaped
			 * with a back slash.
			 * @param Writer The writer to write to.
			 * @return Whether the writing was successful.
			 */
			bool WriteCorrectedForCfgFile(class Config::Writer * Writer);

		public:

			/**
			 * Returns the boolean value of the parameter.
			 * If the parameter is not of boolean type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @return The boolean value of the parameter.
			 */
			bool GetBoolValue();

			/**
			 * Returns the double float value of the parameter.
			 * If the parameter is not of double float type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @return The double float value of the parameter.
			 */
			double GetDoubleValue();

			/**
			 * Returns the long integer value of the parameter.
			 * If the parameter is not of long integer type,
			 * it is converted in prior. The type is not
			 * changed when doing this.
			 * @return The long integer value of the parameter.
			 */
			long GetLongValue();

			/**
			 * Returns the string value of the parameter.
			 * If the parameter is not of string type,
			 * it is converted in prior, as far as this does
			 * not mean any memory allocation. If the conversion
			 * isn't possible without allocating memory (for
			 * example for the integer type), the method
			 * returns an empty string.
			 * @return The string value of the parameter.
			 * @note Do not free or change this value.
			 * Use SetStringValue instead.
			 */
			char * GetStringValue();

			/**
			 * Returns the type of the parameter.
			 * @return The type of the parameter.
			 */
			inline int GetType();

			#ifdef _TOOLBOX_TEST
				/**
				 * Performs a series of tests to check if the class
				 * runs properly. Produces standard output on errors.
				 * @param performedTests Is increased by one for each performed test.
				 * @param failedTests Is increased by one for each failed test.
				 */
				static void RunTestSuite(int * performedTests, int * failedTests);
			#endif

			/**
			 * Sets the type of the parameter to boolean
			 * and sets its value.
			 * @param boolValue The new boolean value.
			 */
			void SetBoolValue(bool boolValue);

			/**
			 * Sets the type of the parameter to double float
			 * and sets its value.
			 * @param doubleValue The new double float value.
			 */
			void SetDoubleValue(double doubleValue);

			/**
			 * Sets the type of the parameter to long integer
			 * and sets its value.
			 * @param longValue The new long integer value.
			 */
			void SetLongValue(long longValue);

			/**
			 * Sets the type of the parameter to string
			 * and sets its value.
			 * @param stringValue The new boolean value.
			 * @note The method copies the string so its memory
			 * is not used after the creation.
			 */
			void SetStringValue(const char * stringValue);
	};
}


#endif
